// id $Id: gguis.c 1.10.1.9 Fri, 05 Sep 2003 22:59:34 +0200 basile $
//  Copyright  2003 Basile STARYNKEVITCH
// $ProjectHeader: Guis 1.51 Fri, 05 Sep 2003 23:19:43 +0200 basile $

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>
#include <time.h>
#include <ctype.h>

#include <glib-object.h>
#include <glib.h>
#include <glib/gmarkup.h>

#include <gtk/gtk.h>

#include <pango/pango.h>


#include "gguis.h"

// put_cstring implemented in common.c
void put_cstring (FILE * f, const char *msg, const char *s, int len);

// initialize the interpreter 
void initialize_interpreter (void);

// load_initial_script implemented in client stuff - return 0 iff ok or an error string
char *load_initial_script (const char *script);

// interpret a request - return 0 iff ok or a static char* error string
char *interpret_request (const char *req);

// run the end of input hook (timeout in milliseconds) return 0 iff ok or a static char* error string
char *end_of_input_hook (int timeout);

#ifndef NDEBUG
void debug_extra (void);

#define dbgprintf(Fmt,Args...) do { if(dbgflag) {			\
  fprintf(stderr, "%s:%d!! " Fmt "\n", __FILE__, __LINE__, ##Args);	\
  fflush(stderr);							\
} } while (0)

#define dbgputs(Msg,Str) do{if (dbgflag) {		\
  fprintf(stderr, "%s:%d!! ", __FILE__, __LINE__);	\
  put_cstring(stderr,(Msg),(Str),-1);  			\
  fflush(stderr);					\
}} while(0)

#define dbgputslen(Msg,Str,Len) do{if (dbgflag) {	\
  fprintf(stderr, "%s:%d!! ", __FILE__, __LINE__);	\
  put_cstring(stderr,(Msg),(Str),(Len));  		\
  fflush(stderr);					\
}} while(0)

#else
#define dbgprintf(Fmt,Args...) do {} while(0)
#define dbgputs(Msg,Str) do {} while(0)
#define dbgputslen(Msg,Str) do {} while(0)
#endif


extern void guis_panic_at (int err, const char *fil, int lin, const char *fct,
                           const char *fmt, ...)
  __attribute__ ((noreturn, format (printf, 5, 6)));

#define guis_panic(Fmt,Args...) \
  guis_panic_at(0, __FILE__, __LINE__, __FUNCTION__, Fmt, ##Args)

#define guis_epanic(Fmt,Args...) \
  guis_panic_at(errno, __FILE__, __LINE__, __FUNCTION__, Fmt, ##Args)

void parse_request_string (char *buf);
void send_reply_string (const char *buf);
void send_reply_printf (const char *fmt, ...)
  __attribute__ ((format (printf, 1, 2)));

void free_gstring (GString * s);

int infd = -1;                  /* input file descriptor */
int outfd = -1;                 /* output file descriptor */
volatile pid_t pipepid = 0;     /* process id of piped
                                   command, cleared by
                                   SIGCHLD signal */
char *pipecmd = 0;              /* piped command */

char *initialscript = 0;        /*  initial script */
char *logfilename = 0;          /* logfilename */
FILE *logfile = 0;              /* logfile for processed requests */
GString *readstr = 0;           /* read buffer string  */
int readoff = 0;                /* offet of last handled byte in read buffer */


int dbgflag = 0;                /* debug flag */
int traceflag = 0;              /* trace flag */
guint inid, outid, idleid;      /* id for input, output and idle callbacks */

GQueue *request_queue;          /* request queue */
GQueue *reply_queue;            /* reply queue */

int nbreq = 0;                  /* request number - seen by script */
int nbsend = 0;                 /* sent reply number - seen by script */
int endtimeout = 0;             /* timeout (in millisecond) - seen & writable by script */

int pipecheckperiod = 1000;     /* period in milliseconds for checking the pipepid */
int pipecheckid;                /*  id for pipecheck perioding callback */

char *progarg0 = "?gguis?";

struct {                        /* trace widgets */
  GtkWidget *topwin;            /* toplevel window */
  GtkWidget *vbox;              /* top vertical box */
  GtkWidget *menubar;           /* the menubar */
  GtkWidget *tracemenu;         /* the trace menu */
  GtkWidget *tracemenuitem;     /* the trace item */
  GtkWidget *traceonitem;       /* the trace on item */
  GtkWidget *traceoffitem;      /* the trace off item */
  GtkWidget *traceclearitem;    /* the trace clear item */
#ifndef NDEBUG
  GtkWidget *debugmenu;         /* the debug menu */
  GtkWidget *debugmenuitem;     /* the debug item */
  GtkWidget *debugonitem;       /* debug on item */
  GtkWidget *debugoffitem;      /* debug off item */
  GtkWidget *debugxtraitem;     /* debug extra item */
#endif
  GtkWidget *label;             /* label */
  GtkTextTagTable *tagtbl;      /* tag table for trace */
  GtkTextTag *tg_tim;           /* tag for time display */
  GtkTextTag *tg_title;         /* tag for title display */
  GtkTextTag *tg_imp;           /* tag for important display */
  GtkTextTag *tg_in;            /* tag for input display */
  GtkTextTag *tg_out;           /* tag for output display */
  GtkTextBuffer *txtbuf;        /* the textbuffer for trace */
  GtkWidget *txtscroll;         /* the scrollbox for trace */
  GtkWidget *txtview;           /* the textview for trace */
} trw = {
0};

const char guis_release[] =
// for PRCS format
// $Format: " \"guis-$ReleaseVersion$\"; "$
 "guis-1.4"; 

static gboolean parse_requests (void *data);
static void insert_trace_time (int dated);
static gboolean end_of_input (void *data);



static void
notify_end (void)
{
  static int endid;
  dbgprintf ("notify_end endid=%d", endid);
  if (!endid) {
    endid = gtk_idle_add_priority (GTK_PRIORITY_LOW, end_of_input, (void *) 0);
    dbgprintf ("added low priority end_of_input endid=%d", endid);
  };
}
static int
pipe_periodical_checker_cb (void *data)
{
  dbgprintf ("pipe checker pipepid=%d", (int) pipepid);
  if (pipepid <= 0) {
    dbgprintf ("pipe checker found dead pid");
    notify_end ();
    return FALSE;
  }
  return TRUE;
}

/* process data arriving on pipe */
static void
reader (gpointer ignored, gint source, GdkInputCondition condition)
{
  int cnt = 0, eof = 0, again = 0;
  GError *err = 0;
  char *eol2 = 0, *eop = 0, *cur = 0, *end = 0;
  GString *str = 0;
  dbgprintf ("reader start source=%d", source);
  again = 1;
  if (infd < 0) {
    dbgprintf ("reader removing inid=%d", inid);
    gtk_input_remove (inid);
    eof = 1;
    again = 0;
  };
  cnt = 0;
  err = 0;
  while (again) {
    int readlen = 0;
    dbgprintf ("startloop readoff=%d readstring@%p ", readoff, readstr);
    if (!readstr)
      readstr = g_string_sized_new (1000);
    dbgprintf ("readstr@%p len%d readoff%d", readstr,
               readstr ? (readstr->len) : 0, readoff);
    assert (readoff >= 0 && readstr && readoff <= readstr->len);
    dbgprintf ("readstr@%p <str@%p len%d alloclen%d>",
               readstr, readstr->str, readstr->len, readstr->allocated_len);
    readlen = readstr->len;
    if (readstr->len + 500 > readstr->allocated_len) {
      int newlen = (5 * readlen) / 4 + 800;
      dbgprintf ("newlen=%d", newlen);
      g_string_set_size (readstr, newlen);
      readstr->len = readlen;
      readstr->str[readlen] = 0;
      dbgprintf ("grown readstr@%p <str@%p len%d alloclen%d>",
                 readstr, readstr->str, readstr->len, readstr->allocated_len);
    }
    readlen = readstr->allocated_len - readstr->len - 3;
    assert (readlen > 100);
    errno = 0;
    dbgprintf ("before read readstr@%p <str@%p len%d alloclen%d> readlen=%d",
               readstr, readstr->str, readstr->len, readstr->allocated_len,
               readlen);
    cnt = read (infd, readstr->str + readstr->len, readlen);
    dbgprintf ("after read cnt=%d error=%s", cnt, strerror (errno));
    if (cnt < 0) {
      if (errno == EINTR)
        continue;
      else
        (again = 0);
    } else if (cnt == 0) {
      eof = 1;
      again = 0;
    } else {
      dbgputslen ("guis reader got", readstr->str + readstr->len, cnt);
      readstr->len += cnt;
      readstr->str[readstr->len] = 0;
      again = 1;
    };
    dbgprintf ("endloop readstr@%p <str@%p len%d alloclen%d> readoff=%d",
               readstr, readstr->str, readstr->len, readstr->allocated_len,
               readoff);
  };                            /* end while again */
  assert (readoff >= 0 && readstr && readoff <= readstr->len);
  if (eof) {
    close (infd);
    infd = -1;
    dbgprintf ("reader eof removing inid=%d", inid);
    gtk_input_remove (inid);
    g_string_append_c (readstr, '\n');
    notify_end ();
  };
  cur = readstr->str + readoff;
  do {
    if (!cur || !cur[0])
      break;
    end = 0;
    eol2 = strstr (cur, "\n\n");
    eop = strchr (cur, '\f');
    if (eol2) {
      if (eop && eop < eol2) {
        end = eop;
        *eop = '\n';
      } else
        end = eol2 + 1;
    } else if (eop) {
      *eop = '\n';
      end = eop;
    } else
      end = 0;
    dbgputs ("cur=", cur);
    dbgprintf ("cur@%p eol2@%p eop@%p end@%p", cur, eol2, eop, end);
    if (end) {
      int siz = 0;
      siz = end - cur + 1;
      while (siz > 0 && (cur[siz - 1] == '\n' || cur[siz - 1] == '\f'))
        siz--;
      dbgprintf ("with end @%p siz%d", end, siz);
      dbgputslen ("new requeststring", cur, siz);
      str = g_string_new_len (cur, siz);
      g_queue_push_head (request_queue, str);
      cur = end + 1;
    };
  } while (end);
  dbgprintf ("cur@%p end@%p", cur, end);
  assert (readoff >= 0 && readstr && readoff <= readstr->len);
  if (cur) {
    readoff = cur - readstr->str;
    assert (readoff <= readstr->len);
    dbgprintf ("readoff=%d cur@%p", readoff, cur);
    dbgprintf ("readstr@%p <str@%p len%d alloclen%d>",
               readstr, readstr->str, readstr->len, readstr->allocated_len);
    if (readoff > 500 + readstr->allocated_len / 4) {
      int newsiz = 5 * (readstr->len - readoff) / 4 + 300;
      GString *newstr = g_string_sized_new (newsiz);
      dbgprintf ("newstr@%p newsiz%d", newstr, newsiz);
      g_string_append_len (newstr, readstr->str + readoff,
                           readstr->len - readoff);
      dbgprintf ("free old readstr@%p", readstr);
      g_string_free (readstr, TRUE);
      readstr = newstr;
      readoff = 0;
      dbgprintf ("readstr@%p <str@%p len%d alloclen%d>",
                 readstr, readstr->str, readstr->len, readstr->allocated_len);
    }
  };
  if (!idleid)
    idleid = gtk_idle_add (parse_requests, (void *) 0);
  dbgprintf ("reader end source=%d idleid=%d", source, idleid);
}                               // end of reader


/* write data on reply pipe */
static void
writer (gpointer ignored, gint source, GdkInputCondition condition)
{
  GString *str = 0;
  int cnt = 0;
  dbgprintf ("writer start source=%d", source);
  if (outfd < 0) {
    dbgprintf ("writer removing outid=%d", outid);
    gtk_input_remove (outid);
    gtk_main_quit ();
    return;
  };
  do {
    if (g_queue_is_empty (reply_queue)) {
      gtk_input_remove (outid);
      outid = 0;
      return;
    };
    str = g_queue_peek_head (reply_queue);
    if (!str)
      return;
    errno = 0;
    dbgputslen ("writing str", str->str, str->len);
    cnt = 0;
    if (str->len > 0) {
      cnt = write (outfd, str->str, str->len);
      dbgprintf ("after write cnt=%d error %s", cnt, strerror (errno));
    };
    if (cnt > 0) {
      if (cnt >= (int) str->len) {
        g_queue_pop_head (reply_queue);
        free_gstring (str);
        str = 0;
      } else
        (void) g_string_erase (str, 0, cnt);
    } else if (cnt < 0 && errno == EPIPE) {
      dbgprintf ("got EPIPE on outfd%d", outfd);
      close (outfd);
      outfd = 0;
      gtk_input_remove (outid);
      notify_end ();
      return;
    }
  } while (cnt > 0);
  dbgprintf ("writer end source=%d idleid=%d", source, idleid);
}                               // end of writer


static void
trace_reply (GString * str)
{
  char bufn[48];
  if (!traceflag)
    return;
  GtkTextIter itend;
  insert_trace_time (0);
  snprintf (bufn, sizeof (bufn), " ?%d ", nbsend);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  gtk_text_buffer_insert (trw.txtbuf, &itend, bufn, -1);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, str->str, str->len,
                                    trw.tg_out, 0);
  if (str->len > 0 && str->str[str->len - 1] != '\n')
    gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
}

void
send_reply_string (const char *buf)
{
  GString *str = 0;
  dbgprintf ("send reply %s outfd=%d", buf, outfd);
  if (!outid) {
    if (outfd >= 0)
      outid =
        gtk_input_add_full (outfd, GDK_INPUT_WRITE, writer, NULL, NULL, NULL);
  };
  str = g_string_new (buf);
  if (str->len <= 0 || str->str[str->len - 1] != '\n')
    str = g_string_append_c (str, '\n');
  g_queue_push_tail (reply_queue, str);
  nbsend++;
  if (traceflag)
    trace_reply (str);
}                               // end of send_reply_string

void
send_reply_printf (const char *fmt, ...)
{
  static char buf[128];
  va_list args;
  int len = 0;
  char *dbuf = 0;
  GString *str = 0;
  memset (buf, 0, sizeof (buf));
  if (!outid) {
    if (outfd >= 0)
      outid =
        gtk_input_add_full (outfd, GDK_INPUT_WRITE, writer, NULL, NULL, NULL);
  };
  va_start (args, fmt);
  len = vsnprintf (buf, sizeof (buf) - 1, fmt, args);
  va_end (args);
  buf[sizeof (buf) - 1] = 0;
  if (len < (int) sizeof (buf) - 2) {
    str = g_string_new (buf);
  } else {
    dbuf = g_malloc0 (len + 2);
    va_start (args, fmt);
    vsnprintf (dbuf, len + 1, fmt, args);
    str = g_string_new (dbuf);
    va_end (args);
    g_free (dbuf);
  };
  if (str->len <= 0 || str->str[str->len - 1] != '\n')
    str = g_string_append_c (str, '\n');
  g_queue_push_tail (reply_queue, str);
  nbsend++;
  if (traceflag)
    trace_reply (str);
}                               // end of send_reply_printf

static gboolean
parse_requests (void *ignored)
{
  GString *str = 0;
  int nb = 0;
  dbgprintf ("begin parse_requests queue=%p %s empty", request_queue,
             g_queue_is_empty (request_queue) ? "is" : "not");
  while (!g_queue_is_empty (request_queue)) {
    nb++;
    dbgprintf ("parse_requests loop %d", nb);
    str = g_queue_peek_tail (request_queue);
    if (str) {
      dbgputs ("parse_requests req", str->str);
      parse_request_string (str->str);
      free_gstring (str);
    } else
      dbgprintf ("request str nil");
    g_queue_pop_tail (request_queue);
    if (nb > 10) {
      dbgprintf ("parse_requests did %d loop and return", nb);
      return TRUE;
    }
  };
  dbgprintf ("end parse_requests %d", nbreq);
  idleid = 0;
  return FALSE;
}                               // end of parse_requests

static int endinputid = 0;
static gboolean
end_of_input (void *data)
{
  char *stat = 0;
  dbgprintf ("end of input endtimeout=%d endinputid=%d", endtimeout,
             endinputid);
  if (endtimeout >= 0) {
    dbgprintf ("running end_input_hook");
    if ((stat = end_of_input_hook (endtimeout)) != 0) {
      fprintf (stderr, "%s - failed to run end_input_hook - %s\n", progarg0,
               stat);
      fflush (stderr);
    };
    if (!endinputid) {
      dbgprintf ("adding main_quit endtimeout %d", endtimeout);
      endinputid = gtk_timeout_add (endtimeout, (GtkFunction) gtk_main_quit, 0);
    };
  } else {
    gtk_main_quit ();
  };
  return FALSE;
}                               // end of end_of_input

static void
usage (char *argv0, int exitcode)
{
  fprintf (stderr,
           "usage: %s [-i inputchan] [-o outputchan]\n"
           "\t [-s initialscript]\n"
           "\t [-L logfile] #log of input requests\n"
           "\t [-v] #show version number\n"
           "\t [-T] #show protocol trace window (or GGUIS_TRACE env.var =1)\n"
           "\t [-p pipecommand]\n" "\t [...gtkargs]\n"
#ifndef NDEBUG
           "\t [-D] #for debug output (or GGUIS_DEBUG env.var =1)\n"
#endif
           "\t [-h] #for this help\n"
           "where a channel is either a file descriptor number or a file path\n"
           "eg -i 0 or -i - reads from stdin and -i /tmp/input_pipe reads from a namedpipe \n"
           "and the pipecommand is a command whose stdin is the inputchan\n"
           "and whose stdout is the outputchan\n"
           "if command name ends with '-scripter' then the next argument is the initial script\n"
           "[this release is %s]\n\n%s\n", argv0, guis_release, guis_doc);
  if (exitcode >= 0)
    exit (exitcode);
}



// utility to append the XML repr of a unicode char
void
guis_append_gstring_xml_unichar (GString * gs, int c)
{
  switch (c) {
  case 'a'...'z':
  case 'A'...'Z':
  case '0'...'9':
  case ' ':
  case '+':
  case '-':
  case '(':
  case ')':
    g_string_append_c (gs, c);
    break;
  case '&':
    g_string_append (gs, "&amp;");
    break;
  case '\'':
    g_string_append (gs, "&apos;");
    break;
  case '<':
    g_string_append (gs, "&lt;");
    break;
  case '"':
    g_string_append (gs, "&quot;");
    break;
  case '>':
    g_string_append (gs, "&gt;");
    break;
  default:
    if (c > ' ' && (int) c < 127 && isprint (c))
      g_string_append_c (gs, c);
    else
      g_string_append_printf (gs, "&#%d;", (int) c & 0xffffff);
    break;
  }
}

void
guis_append_gstring_cguis_unichar (GString * gs, int c)
{
  switch (c) {
  case 'a'...'z':
  case 'A'...'Z':
  case '0'...'9':
  case ' ':
  case '_':
  case '+':
  case '-':
  case '*':
  case '/':
  case '&':
  case '<':
  case '>':
    g_string_append_c (gs, c);
    break;
  case '\n':
    g_string_append (gs, "\\n");
    break;
  case '\t':
    g_string_append (gs, "\\t");
    break;
  case '\f':
    g_string_append (gs, "\\f");
    break;
  case '\v':
    g_string_append (gs, "\\v");
    break;
  case '\r':
    g_string_append (gs, "\\r");
    break;
  case '\'':
    g_string_append (gs, "\\'");
    break;
  case '\"':
    g_string_append (gs, "\\Q");
    break;
  default:
    if (c > 32 && c < 127 && isprint (c))
      g_string_append_c (gs, c);
    else if (c >= 0 && c < 256)
      g_string_append_printf (gs, "\\x%02x", ((int) c) & 0xff);
    else
      g_string_append_printf (gs, "\\u%04x", c);
    break;
  };
}

static void
parseargs (int argc, char **argv)
{
  int i = 0;
  int inpipefds[2] = {
    -1, -1
  };
  int outpipefds[2] = {
    -1, -1
  };
  char *inarg = 0;
  char *outarg = 0;
  char *pipearg = 0;
  for (i = 1; i < argc; i++) {
    char *curarg = argv[i];
    if (curarg[0] == '-') {
#define ISLONGARG(Arg) (curarg[1]=='-' && !strcmp(curarg,Arg))
      if (curarg[1] == 'i' || ISLONGARG ("--input")) {
        if (curarg[2])
          inarg = curarg + 2;
        else if (i < argc - 1)
          inarg = argv[++i];
      } else if (curarg[1] == 'o' || ISLONGARG ("--output")) {
        if (curarg[2])
          outarg = curarg + 2;
        else if (i < argc - 1)
          outarg = argv[++i];
      } else if (curarg[1] == 'p' || ISLONGARG ("--pipe")) {
        if (curarg[2])
          pipearg = curarg + 2;
        else if (i < argc - 1)
          pipearg = argv[++i];
      } else if (curarg[1] == 's' || ISLONGARG ("--script")) {
        if (curarg[2])
          initialscript = curarg + 2;
        else if (i < argc - 1)
          initialscript = argv[++i];
      } else if (curarg[1] == 'L' || ISLONGARG ("--logfile")) {
        if (curarg[2])
          logfilename = curarg + 2;
        else if (i < argc - 1)
          logfilename = argv[++i];
      } else if (curarg[1] == 'h' || ISLONGARG ("--help"))
        usage (argv[0], 0);
      else if (curarg[1] == 'T' || ISLONGARG ("--trace"))
        traceflag = 1;
#ifndef NDEBUG
      else if (curarg[1] == 'D' || ISLONGARG ("--debug"))
        dbgflag = 1;
#endif
      else if (curarg[1] == 'v' || ISLONGARG ("--version"))
        fprintf (stderr,
                 "%s:\n release %s\n compiled on " __DATE__ " at " __TIME__
                 "\n", argv[0], guis_release);
      else
        goto badarg;
    } else {
    badarg:
      fprintf (stderr, "%s bad argument %s\n", argv[0], curarg);
      usage (argv[0], 1);
    }
  };
  if (inarg) {
    if (isdigit (inarg[0]))
      infd = atoi (inarg);
    else if (!strcmp (inarg, "-"))
      infd = STDIN_FILENO;
    else {
      infd = open (inarg, O_RDONLY);
      if (infd < 0) {
        fprintf (stderr, "%s bad input %s : %s\n",
                 argv[0], inarg, strerror (errno));
        exit (1);
      }
    }
  }
  if (outarg) {
    if (isdigit (outarg[0]))
      outfd = atoi (outarg);
    else if (!strcmp (outarg, "-"))
      outfd = STDOUT_FILENO;
    else {
      outfd = open (outarg, O_WRONLY);
      if (outfd < 0) {
        fprintf (stderr, "%s bad output %s : %s\n",
                 argv[0], outarg, strerror (errno));
        exit (1);
      }
    }
  }
  if (pipearg) {
    pipecmd = pipearg;
    if (infd >= 0 || outfd >= 0) {
      fprintf (stderr,
               "%s pipe command -p %s incompatible with -i or -o\n",
               argv[0], pipearg);
      exit (1);
    };
    if (pipe (inpipefds)) {
      perror ("pipe in");
      exit (1);
    };
    if (pipe (outpipefds)) {
      perror ("pipe out");
      exit (1);
    };
    pipepid = fork ();
    if (pipepid == 0) {
      /* child process */
      int i = 0;
      char *shell = 0;
      nice (2);
      if (dup2 (outpipefds[0], 0) < 0 || dup2 (inpipefds[1], 1) < 0) {
        perror ("dup2");
        exit (1);
      };
      close (outpipefds[1]);
      close (inpipefds[0]);
      for (i = 32; i > 2; i--)
        close (i);
      dbgprintf ("before exec %s\n", pipearg);
      if (!strchr (pipearg, ' '))
        /* optimize case when no space, so no shell needed... */
        execlp (pipearg, pipearg, 0);
      else {
        shell = getenv ("SHELL");
        if (!shell)
          shell = "/bin/sh";
        execl (shell, shell, "-c", pipearg, 0);
      }
      fprintf (stderr, "exec %s failed - %s\n", pipearg, strerror (errno));
      exit (1);
    } else if (pipepid < 0) {
      perror ("fork");
      exit (1);
    } else {
      /* parent process */
      usleep (5000);            /* sleep a bit to let child run */
      infd = inpipefds[0];
      close (inpipefds[1]);
      outfd = outpipefds[1];
      close (outpipefds[0]);
      dbgprintf ("after fork pipepid=%d infd=%d outfd=%d", pipepid, infd,
                 outfd);
    }
  }
  if (infd >= 0) {
    fcntl (infd, F_SETFL, O_NONBLOCK);
    dbgprintf ("nonblocking infd%d", infd);
  };
  if (outfd >= 0) {
    fcntl (outfd, F_SETFL, O_NONBLOCK);
    dbgprintf ("nonblocking outfd%d", outfd);
  };
}                               // end of parseargs

static void
sigchld_handler (int sig)
{
  int st = 0;
  pid_t pid = 0;
  if (pipepid > 0 && (pid = waitpid (pipepid, &st, WNOHANG)) == pipepid) {
    dbgprintf ("sigchld handler st=%d pid %d", st, (int) pid);
    if (infd >= 0)
      close (infd);
    if (outfd >= 0)
      close (outfd);
    infd = -1;
    outfd = -1;
    pipepid = -1;
    dbgprintf ("sigchld handler cleared pipepid pid%d", pid);
  };
}                               // end of sigchld_handler


static void
insert_trace_time (int dated)
{
  struct {
    char buf[200];
    GTimeVal tv;
    struct tm tm;
    char sec[10];
  } t;
  time_t tim;
  GtkTextIter itend;
  memset (&t, 0, sizeof (t));
  g_get_current_time (&t.tv);
  tim = t.tv.tv_sec;
  localtime_r (&tim, &t.tm);
  strftime (t.buf, sizeof (t.buf) - 10, dated ? "%G %b %d @ %T" : "%T", &t.tm);
  sprintf (t.sec, ".%03d", (int) t.tv.tv_usec / 1000);
  strcat (t.buf, t.sec);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, t.buf, -1,
                                    trw.tg_tim, 0);
}

static void
traceon_cb (void *unused)
{
  GtkTextIter itend;
  char buf[48];
  dbgprintf ("traceon");
  insert_trace_time (1);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  snprintf (buf, sizeof (buf), "TRACE ON #%d", nbreq);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, buf, -1,
                                    trw.tg_title, 0);
  gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
  traceflag = 1;
}

static void
traceoff_cb (void *unused)
{
  char buf[48];
  GtkTextIter itend;
  dbgprintf ("traceoff");
  insert_trace_time (1);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  snprintf (buf, sizeof (buf), "TRACE OFF #%d", nbreq);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, buf, -1,
                                    trw.tg_title, 0);
  gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
  traceflag = 0;
}

static void
traceclear_cb (void *unused)
{
  GtkTextIter itend;
  char buf[48];
  dbgprintf ("traceclear");
  gtk_text_buffer_set_text (trw.txtbuf, "", 0);
  insert_trace_time (1);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  snprintf (buf, sizeof (buf), "TRACE CLEARED #%d", nbreq);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, buf, -1,
                                    trw.tg_title, 0);
  gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
  traceflag = 0;
}

#ifndef NDEBUG
static void
debugon_cb (void *unused)
{
  GtkTextIter itend;
  char buf[48];
  dbgflag = 1;
  fprintf (stderr, "\ngguis: **** debug on ***\n");
  insert_trace_time (1);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  snprintf (buf, sizeof (buf), "DEBUG ON #%d", nbreq);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, buf, -1,
                                    trw.tg_title, 0);
  gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
}

static void
debugoff_cb (void *unused)
{
  char buf[48];
  GtkTextIter itend;
  fprintf (stderr, "\ngguis: **** debug off ***\n");
  insert_trace_time (1);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  snprintf (buf, sizeof (buf), "DEBUG OFF #%d", nbreq);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, buf, -1,
                                    trw.tg_title, 0);
  gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
  dbgflag = 0;
}

static void
debugxtra_cb (void *unused)
{
  char buf[48];
  GtkTextIter itend;
  fprintf (stderr, "\ngguis: **** debug xtra ***\n");
  insert_trace_time (1);
  gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
  snprintf (buf, sizeof (buf), "DEBUG XTRA #%d", nbreq);
  gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, buf, -1,
                                    trw.tg_title, 0);
  gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
  debug_extra ();
}
#endif


static void
initialize (void)
{
  static char buf[300];
  char hnam[100];
  dbgprintf ("start of initialize compiled %s @%s", __DATE__, __TIME__);
  if (traceflag) {
    trw.topwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (trw.topwin), "guis trace");
    gtk_window_set_default_size (GTK_WINDOW (trw.topwin), 650, 400);
    trw.vbox = gtk_vbox_new (FALSE /*heterogenous */ , 5);
    gtk_container_add (GTK_CONTAINER (trw.topwin), trw.vbox);
    trw.menubar = gtk_menu_bar_new ();
    ////
    trw.tracemenu = gtk_menu_new ();
    trw.tracemenuitem = gtk_menu_item_new_with_label ("Trace");
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.menubar), trw.tracemenuitem);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (trw.tracemenuitem),
                               trw.tracemenu);
    trw.traceonitem = gtk_menu_item_new_with_mnemonic ("trace _on");
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.tracemenu), trw.traceonitem);
    g_signal_connect_swapped (G_OBJECT (trw.traceonitem), "activate",
                              G_CALLBACK (traceon_cb), (gpointer) 0);
    trw.traceoffitem = gtk_menu_item_new_with_mnemonic ("trace o_ff");
    g_signal_connect_swapped (G_OBJECT (trw.traceoffitem), "activate",
                              G_CALLBACK (traceoff_cb), (gpointer) 0);
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.tracemenu), trw.traceoffitem);
    trw.traceclearitem = gtk_menu_item_new_with_mnemonic ("trace _clear");
    g_signal_connect_swapped (G_OBJECT (trw.traceclearitem), "activate",
                              G_CALLBACK (traceclear_cb), (gpointer) 0);
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.tracemenu), trw.traceclearitem);
    ////
#ifndef NDEBUG
    trw.debugmenu = gtk_menu_new ();
    trw.debugmenuitem = gtk_menu_item_new_with_label ("Debug");
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.menubar), trw.debugmenuitem);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (trw.debugmenuitem),
                               trw.debugmenu);
    trw.debugonitem = gtk_menu_item_new_with_mnemonic ("debug _on");
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.debugmenu), trw.debugonitem);
    g_signal_connect_swapped (G_OBJECT (trw.debugonitem), "activate",
                              G_CALLBACK (debugon_cb), (gpointer) 0);
    trw.debugoffitem = gtk_menu_item_new_with_mnemonic ("debug o_ff");
    g_signal_connect_swapped (G_OBJECT (trw.debugoffitem), "activate",
                              G_CALLBACK (debugoff_cb), (gpointer) 0);
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.debugmenu), trw.debugoffitem);
    trw.debugxtraitem = gtk_menu_item_new_with_mnemonic ("debug e_xtra");
    g_signal_connect_swapped (G_OBJECT (trw.debugxtraitem), "activate",
                              G_CALLBACK (debugxtra_cb), (gpointer) 0);
    gtk_menu_shell_append (GTK_MENU_SHELL (trw.debugmenu), trw.debugxtraitem);
#endif
    ////
    gtk_box_pack_start (GTK_BOX (trw.vbox), trw.menubar, /*expand= */ FALSE,
                        /*fill= */ FALSE, /*space= */ 3);
    trw.label = gtk_label_new (0);
    gtk_box_pack_start (GTK_BOX (trw.vbox), trw.label, /*expand= */ FALSE,
                        /*fill= */ FALSE, /*space= */ 3);
    memset (hnam, 0, sizeof (hnam));
    gethostname (hnam, sizeof (hnam) - 1);
    snprintf (buf, sizeof (buf) - 1,
              "<b>%s</b> <tt>%s</tt> pid%d on <tt>%s</tt>\n<small>compiled on "
              __DATE__ "@" __TIME__ "</small>",
              progarg0, initialscript ? : "", (int) getpid (), hnam);
    gtk_label_set_markup (GTK_LABEL (trw.label), buf);
    trw.txtscroll = gtk_scrolled_window_new (0, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (trw.txtscroll),
                                    GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    gtk_box_pack_start (GTK_BOX (trw.vbox), trw.txtscroll, /*expand= */ TRUE,
                        /*fill= */ TRUE, /*space= */ 3);
    trw.tagtbl = gtk_text_tag_table_new ();
    trw.txtbuf = gtk_text_buffer_new (trw.tagtbl);
    trw.tg_tim = gtk_text_buffer_create_tag (trw.txtbuf, "tim",
                                             "weight", PANGO_WEIGHT_BOLD,
                                             "scale", PANGO_SCALE_SMALL,
                                             "foreground", "DarkGoldenrod4", 0);
    trw.tg_title = gtk_text_buffer_create_tag (trw.txtbuf, "title",
                                               "scale", PANGO_SCALE_X_LARGE,
                                               "foreground", "red", 0);
    trw.tg_imp = gtk_text_buffer_create_tag (trw.txtbuf, "imp",
                                             "scale", PANGO_SCALE_LARGE,
                                             "weight", PANGO_WEIGHT_BOLD,
                                             "foreground", "red", 0);
    trw.tg_in = gtk_text_buffer_create_tag (trw.txtbuf, "in",
                                            "weight", PANGO_WEIGHT_BOLD,
                                            "foreground", "blue", 0);
    trw.tg_out = gtk_text_buffer_create_tag (trw.txtbuf, "out",
                                             "style", PANGO_STYLE_ITALIC,
                                             "foreground", "darkgreen", 0);
    trw.txtview = gtk_text_view_new_with_buffer (trw.txtbuf);
    gtk_text_view_set_editable (GTK_TEXT_VIEW (trw.txtview), FALSE);
    gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (trw.txtview), GTK_WRAP_CHAR);
    gtk_container_add (GTK_CONTAINER (trw.txtscroll), trw.txtview);
    {
      GtkTextIter itend;
      insert_trace_time (1);
      memset (buf, 0, sizeof (buf));
      gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
      gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend,
                                        "TRACE WINDOW\n", -1, trw.tg_title, 0);
      gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend,
                                        guis_release, -1, trw.tg_tim, 0);
      snprintf (buf, sizeof (buf) - 1, " for %s\n", guis_language_version ());
      gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend,
                                        buf, -1, trw.tg_tim, 0);
      gtk_text_buffer_insert (trw.txtbuf, &itend,
                              "see http://starynkevitch.net/Basile/guisdoc.html"
                              " or freshmeat.net/projects/guis\n"
                              "input requests shown ", -1);
      gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend,
                                        "like this", -1, trw.tg_in, 0);
      gtk_text_buffer_insert (trw.txtbuf, &itend,
                              " & output replies shown ", -1);
      gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend,
                                        "like that\n\n", -1, trw.tg_out, 0);
    }
    gtk_widget_show_all (trw.topwin);
    //g_signal_connect (G_OBJECT (trw.topwin), "delete_event",
    //                  G_CALLBACK (delete_event), NULL);
  };
  dbgprintf ("end of initialize");
}                               // end of initialize



int
main (int argc, char **argv)
{
  const char *status = 0;
  char *envar = 0;
  struct sigaction sac;
  // horrible trick to be able to accept script starting with
  // #! /usr/bin/env *-scripter
  progarg0 = argv[0];
  {
    int argv0len = 0;
    if (argv[0] && (argv0len = strlen (argv[0])) >= 10
        && !strcmp (argv[0] + argv0len - 9, "-scripter")) {
      if (argc > 1 && !access (argv[1], R_OK)) {
        initialscript = argv[1];
        argv += 1;
        argc -= 1;
      }
    }
  }
#ifndef NDEBUG
  envar = getenv ("GGUIS_DEBUG");
  if (envar) {
    dbgflag = atoi (envar);
    fprintf (stderr, "GGUIS_DEBUG sets dbgflag to %d\n", dbgflag);
  }
#endif
  envar = getenv ("GGUIS_TRACE");
  if (envar)
    traceflag = atoi (envar);
  dbgprintf ("before gtk_init argc=%d", argc);
  gtk_init (&argc, &argv);
  dbgprintf ("after gtk_init argc=%d", argc);
  memset (&sac, 0, sizeof (sac));
  sac.sa_handler = sigchld_handler;
  sac.sa_flags = SA_NOCLDSTOP | SA_RESTART;
  if (sigaction (SIGCHLD, &sac, (void *) 0))
    perror ("sigaction");
  parseargs (argc, argv);
  if (logfilename) {
    logfile = fopen (logfilename, "w");
    dbgprintf ("logfilename %s logfile %p #%d", logfilename, logfile,
               logfile ? fileno (logfile) : (-1));
    if (logfile)
      fprintf (stderr, "%s logging input requests on %s\n", progarg0,
               logfilename);
    else
      fprintf (stderr, "%s cannot open logfile %s - %s\n",
               progarg0, logfilename, strerror (errno));
  };
  dbgprintf ("gguis pid %d compiled " __DATE__ "@" __TIME__ "\n",
             (int) getpid ());
  initialize ();
  initialize_interpreter ();
  dbgputs ("loading script", initialscript);
  status = load_initial_script (initialscript);
  dbgputs ("initial script status", status);
  if (status) {
    fprintf (stderr, "gguis failed to initialize %s - %s\n",
             initialscript ? : "-", status);
    exit (EXIT_FAILURE);
  }
  request_queue = g_queue_new ();
  reply_queue = g_queue_new ();
  if (infd >= 0)
    inid = gtk_input_add_full (infd, GDK_INPUT_READ, reader, NULL, NULL, NULL);
#ifndef NDEBUG
  if (dbgflag && pipecheckperiod > 0)
    pipecheckperiod += pipecheckperiod * 2 + 3000;
#endif
  if (pipepid > 0 && pipecheckperiod > 0) {
    pipecheckid = gtk_timeout_add (pipecheckperiod,
                                   pipe_periodical_checker_cb, (void *) 0);
    dbgprintf ("pipecheckid %d pipepid %d period %d",
               pipecheckid, (int) pipepid, pipecheckperiod);
  }
  if (script_without_main_loop ()) {
    dbgprintf ("guis pid %d don't call gtk_main", (int) getpid ());
  } else {
    dbgprintf ("guis before gtk_main pid%d", (int) getpid ());
    gtk_main ();
  }
  if (logfile) {
    fclose (logfile);
    logfile = 0;
  };
  dbgprintf ("guis ended pid%d", (int) getpid ());
  return 0;
}                               // end of main

void
set_pipe_check_period (int newper)
{
  int oldper = pipecheckperiod;
  if (newper > 0 && newper < 100)
    newper = 100;
  else if (newper > 10000)
    newper = 10000;
  if (oldper != newper) {
    if (pipecheckid > 0)
      gtk_timeout_remove (pipecheckid);
    if (pipepid > 0 && newper > 0)
      pipecheckid = gtk_timeout_add (pipecheckperiod,
                                     pipe_periodical_checker_cb, (void *) 0);
    pipecheckperiod = newper;
  }
}

//////////////////////////////////////////////////////////////////
// usual panic function
void
guis_panic_at (int err, const char *fil, int lin, const char *fct,
               const char *fmt, ...)
{
  va_list ap;
  fflush (stderr);
  fflush (stdout);
  fprintf (stderr, "\n**GUIS PANIC [%s:%d", fil, lin);
  if (fct && fct[0])
    fprintf (stderr, "@%s]", fct);
  else
    putc (']', stderr);
  putc (' ', stderr);
  fflush (stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  if (err > 0)
    fprintf (stderr, ":%d - %s", err, strerror (err));
  putc ('\n', stderr);
  fflush (stderr);
  abort ();
   /*NOTREACHED*/ exit (2);
}                               // end of guis_panic_at

void
free_gstring (GString * str)
{
  if (!str)
    return;
  memset (str->str, 0, str->len);
  str->len = 0;
  g_string_free (str, TRUE);
}                               // end of free_gstring 




void
parse_request_string (char *str)
{
  char bufn[48];
  char bufe[256];
  char *status = 0;
  GtkTextIter itend;
  nbreq++;
  dbgprintf ("request#%d", nbreq);
  dbgputs ("process request", str);
  if (traceflag) {
    int slen = strlen (str);
    insert_trace_time (0);
    gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
    memset (bufn, 0, sizeof (bufn));
    snprintf (bufn, sizeof (bufn) - 1, " #%d ", nbreq);
    gtk_text_buffer_insert (trw.txtbuf, &itend, bufn, -1);
    gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, str, -1,
                                      trw.tg_in, 0);
    if (slen > 0 && str[slen - 1] != '\n')
      gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", 1);
  };
  if (logfile) {
    dbgprintf ("logging request");
    fputs (str, logfile);
    putc ('\n', logfile);
    putc ('\n', logfile);
    fflush (logfile);
  }
  status = interpret_request (str);
  if (status) {
    fprintf (stderr,
             "gguis - failed to execute request %.120s : %s\n", str, status);
    if (traceflag) {
      memset (bufe, 0, sizeof (bufe));
      gtk_text_buffer_get_end_iter (trw.txtbuf, &itend);
      snprintf (bufe, sizeof (bufe) - 1, "req.#%d failed: ", nbreq);
      gtk_text_buffer_insert_with_tags (trw.txtbuf, &itend, bufe, -1,
                                        trw.tg_imp, 0);
      gtk_text_buffer_insert (trw.txtbuf, &itend, status, -1);
      gtk_text_buffer_insert (trw.txtbuf, &itend, "\n", -1);
    }
  }
}                               /* end of parse_request_string */

/* eof $Id: gguis.c 1.10.1.9 Fri, 05 Sep 2003 22:59:34 +0200 basile $ */
